//////////////////////////////////////////////////////////////////////
// Simple ARM7 stub (sends RTC, TSC, and X/Y data to the ARM 9)
// -- joat
// -- modified by Darkain and others
//////////////////////////////////////////////////////////////////////

#include "../../ipc.h"
#include <NDS.h>
#include "emu2149.h"
#include "emu2212.h"

//////////////////////////////////////////////////////////////////////


#define TOUCH_CAL_X1 (*(vs16*)0x027FFCD8)
#define TOUCH_CAL_Y1 (*(vs16*)0x027FFCDA)
#define TOUCH_CAL_X2 (*(vs16*)0x027FFCDE)
#define TOUCH_CAL_Y2 (*(vs16*)0x027FFCE0)
#define SCREEN_WIDTH    256
#define SCREEN_HEIGHT   192
s32 TOUCH_WIDTH  = TOUCH_CAL_X2 - TOUCH_CAL_X1;
s32 TOUCH_HEIGHT = TOUCH_CAL_Y2 - TOUCH_CAL_Y1;
s32 TOUCH_OFFSET_X = ( ((SCREEN_WIDTH -60) * TOUCH_CAL_X1) / TOUCH_WIDTH  ) - 28;
s32 TOUCH_OFFSET_Y = ( ((SCREEN_HEIGHT-60) * TOUCH_CAL_Y1) / TOUCH_HEIGHT ) - 28;

PSG * psg;
SCC * scc;
#define MSX_CLK 3579545
#define SAMPLERATE  22050  //22050
#define SOUNDBUFCNT ((u16)(SAMPLERATE/5))   //367

#define TIMERCOUNT  ( 33554432 / SAMPLERATE )

u16 soundbuf[4][SOUNDBUFCNT];

int pbufp = 0; // f[^vCobt@
int bufcnt= 0; // ܂ꂽobt@

#define SCMD_PSG_Reset     1
#define SCMD_PSG_WriteReg  2
#define SCMD_PSG_readReg   3
#define SCMD_SCC_Reset     4
#define SCMD_SCC_WriteReg  5

//////////////////////////////////////////////////////////////////////

void startSound(int sampleRate, const void* data, uint32 bytes, u8 channel=0, u8 vol=0x7F,  u8 pan=63, u8 format=0) {
  SCHANNEL_CR(channel)     = 0;
  SCHANNEL_TIMER(channel)  = SOUND_FREQ(sampleRate);
  SCHANNEL_SOURCE(channel) = (uint32)data;
  SCHANNEL_LENGTH(channel) = bytes >> 2;
  SCHANNEL_REPEAT_POINT(channel) = 0;
  SCHANNEL_CR(channel)     = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(vol) | SOUND_PAN(pan) | (format==1? SOUND_FORMAT_8BIT:SOUND_FORMAT_16BIT);
}

bool isFreeSoundChannel(int i) {
  if ( (SCHANNEL_CR(i) & SCHANNEL_ENABLE) == 0 ) return true;
  return false;
}

//////////////////////////////////////////////////////////////////////

void InitSound(int r)
{
	// PSG initialize
	psg = PSG_new(MSX_CLK,r);
	PSG_reset( psg );

	// SCC initialize
	scc = SCC_new(MSX_CLK,r);
	SCC_reset( scc );
	
	IPC->soundCommandResult	= 1;
}

void vsync()
{
	static int heartbeat = 0;
 
	uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0, batt=0, aux=0;
	int t1=0, t2=0;
	uint32 temp=0;
	uint8 ct[sizeof(IPC->curtime)];

	// Update the heartbeat
	heartbeat++;

	// Read the X/Y buttons and the /PENIRQ line
	but = XKEYS;
	if (!(but & 0x40)) {
		// Read the touch screen
		x = touchRead(TSC_MEASURE_X);
		y = touchRead(TSC_MEASURE_Y);
		xpx = ( ((SCREEN_WIDTH -60) * x) / TOUCH_WIDTH  ) - TOUCH_OFFSET_X;
		ypx = ( ((SCREEN_HEIGHT-60) * y) / TOUCH_HEIGHT ) - TOUCH_OFFSET_Y;
		z1 = touchRead(TSC_MEASURE_Z1);
		z2 = touchRead(TSC_MEASURE_Z2);
	}

	batt = touchRead(TSC_MEASURE_BATTERY);
	aux  = touchRead(TSC_MEASURE_AUX);

	// Read the time
	rtcGetTime((uint8 *)ct);
	BCDToInteger((uint8 *)&(ct[1]), 7);

	// Read the temperature
	temp = touchReadTemperature(&t1, &t2);

	// Update the IPC struct
	IPC->heartbeat = heartbeat;
	IPC->buttons   = but;
	IPC->touchX    = x;
	IPC->touchY    = y;
	IPC->touchXpx  = xpx;
	IPC->touchYpx  = ypx;
	IPC->touchZ1   = z1;
	IPC->touchZ2   = z2;
	IPC->battery   = batt;
	IPC->aux       = aux;

	for(u32 i=0; i<sizeof(ct); i++) {
	  IPC->curtime[i] = ct[i];
	}

	IPC->temperature = temp;
	IPC->tdiode1 = t1;
	IPC->tdiode2 = t2;
//	VBLANK_INTR_WAIT_FLAGS |= IRQ_VBLANK;
//	IF |= IRQ_VBLANK;
}

void InterruptHandler(void) {
/*
	static int heartbeat = 0;
 
	if (IF & IRQ_VBLANK) {
 		uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0, batt=0, aux=0;
		int t1=0, t2=0;
		uint32 temp=0;
		uint8 ct[sizeof(IPC->curtime)];

    
		// Update the heartbeat
		heartbeat++;

		// Read the X/Y buttons and the /PENIRQ line
		but = XKEYS;
		if (!(but & 0x40)) {
			// Read the touch screen
			x = touchRead(TSC_MEASURE_X);
			y = touchRead(TSC_MEASURE_Y);
			xpx = ( ((SCREEN_WIDTH -60) * x) / TOUCH_WIDTH  ) - TOUCH_OFFSET_X;
			ypx = ( ((SCREEN_HEIGHT-60) * y) / TOUCH_HEIGHT ) - TOUCH_OFFSET_Y;
			z1 = touchRead(TSC_MEASURE_Z1);
			z2 = touchRead(TSC_MEASURE_Z2);
		}

		batt = touchRead(TSC_MEASURE_BATTERY);
		aux  = touchRead(TSC_MEASURE_AUX);

		// Read the time
		rtcGetTime((uint8 *)ct);
		BCDToInteger((uint8 *)&(ct[1]), 7);

		// Read the temperature
		temp = touchReadTemperature(&t1, &t2);

		// Update the IPC struct
		IPC->heartbeat = heartbeat;
		IPC->buttons   = but;
		IPC->touchX    = x;
		IPC->touchY    = y;
		IPC->touchXpx  = xpx;
		IPC->touchYpx  = ypx;
		IPC->touchZ1   = z1;
		IPC->touchZ2   = z2;
		IPC->battery   = batt;
		IPC->aux       = aux;

		for(u32 i=0; i<sizeof(ct); i++) {
		  IPC->curtime[i] = ct[i];
		}

		IPC->temperature = temp;
		IPC->tdiode1 = t1;
		IPC->tdiode2 = t2;
		VBLANK_INTR_WAIT_FLAGS |= IRQ_VBLANK;
		IF |= IRQ_VBLANK;
	}
	*/
	if (IF & IRQ_TIMER0) {
		if( bufcnt ){
			if( isFreeSoundChannel(0) ){
				startSound(SAMPLERATE, &soundbuf[pbufp], SOUNDBUFCNT*2, 0 , 0x7f, 63, 0);
				pbufp = (pbufp + 1) & 0x3;
				bufcnt--;
			}
		}
		VBLANK_INTR_WAIT_FLAGS |= IRQ_TIMER0;
		IF |= IRQ_TIMER0;
	}
	// Acknowledge interrupts
//	IF = IF;
}
 

/** LoopSound() **********************************************/
/** Main loop of the sound server.                          **/
/*************************************************************/
void ProcSound()
{
	e_uint16 P,S,O,A;
	int soundloop_stat = 1;
	static int wbufp = 0; // f[^obt@
	static int wptr  = 0; // f[^ʒu
	P=O=A=S=0;

	switch( IPC->soundCommand ){
		case SCMD_PSG_Reset :
			PSG_reset(psg);
			break;
		case SCMD_PSG_WriteReg :
			PSG_writeReg(psg,IPC->param1,IPC->param2);
			break;
		case SCMD_PSG_readReg :
			IPC->retValue1 = PSG_readReg(psg,IPC->param1);
			break;
		case SCMD_SCC_Reset :
			SCC_reset(scc);
			break;
		case SCMD_SCC_WriteReg :
			PSG_writeReg(psg,IPC->param1,IPC->param2);
			break;
	};
	IPC->soundCommand = 0;
	IPC->soundCommandResult = 1;

	if( bufcnt < 4 ){
		P = PSG_calc(psg);
		S = SCC_calc(scc);
//			if(snd.opll) O = OPLL_calc(snd.opll);
//			if(snd.opl)  A = OPL_calc(snd.opl);
		soundbuf[wbufp][wptr] = P+S; //+O+(A>>2);

		wptr = (wptr + 1) % SOUNDBUFCNT;
		if( !wptr ){
			wbufp = (wbufp + 1) & 0x3; 
			bufcnt ++;
		}
	}
}

//////////////////////////////////////////////////////////////////////

int snd[8] = { 0x1AC, 0x17d, 0x153, 0x140, 0x11D, 0xFE, 0xE3, 0xD6 };

void playSound()
{
	static int play = false;
	static int cnt = 0;
	static int soundbeat = 0;

	if( !play ){
		PSG_writeReg(psg,7,0xBE);
		PSG_writeReg(psg,8,0x0F);
		play = true;
	}

	soundbeat = (soundbeat+1) & 0xFFFF;

	if( !soundbeat ){
		PSG_writeReg(psg,1,(u8)((snd[cnt]>>8)&0x0F));
		PSG_writeReg(psg,0,(u8)(snd[cnt]&0xFF));
		cnt = (cnt+1) & 7;
	}
	
}
 

int main(int argc, char ** argv) {

	int vsyncflg = true;
	
	// Reset the clock if needed
	rtcReset();

	//enable sound
	powerON(POWER_SOUND);
	SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);

	// Timer setting
	TIMER0_DATA = 0xFFF0;
	TIMER0_CR = TIMER_DIV_64 | TIMER_IRQ_REQ | TIMER_ENABLE;

	// Sound Initialize
	InitSound(SAMPLERATE);

	// Set up the interrupt handler
	IME = 0;
	IRQ_HANDLER = &InterruptHandler;
	IE = IRQ_VBLANK | IRQ_TIMER0;
	DISP_SR = DISP_VBLANK_IRQ;
	IF = ~0;
	IME = 1;

	while (1){
//		playSound();
		ProcSound();
		if( !vsyncflg ){
			if( !(DISP_SR & DISP_IN_VBLANK) ){
				vsync();
				vsyncflg = true;
			}
		}else{
			if( (DISP_SR & DISP_IN_VBLANK) ){
				vsyncflg = false;
			}
		}
	}
	return 0;
}

 
//////////////////////////////////////////////////////////////////////

